#  SiteCost
#
#     2020-03-18  by TAKENAKA, Akio

import os.path
import sys
import re

# J: ファイル入出力時の文字コード指定
CHR_CODE = 'utf-8'  

# J: ####################################
# J: 各サイトのコストを管理するクラス

class SiteCost:
    """
    This class handles cost data of assigning sites to the protected area
    
    """

    def __init__(self):

        # J: コストを記憶する辞書。キーはサイトID、値はコスト
        # Dictionary to contain cost of each site.
        # The keys will be site IDs and the values are the cost.

        self.costs = {}


    # J: ####################################
    # J: コストの計算。指定サイトのコストを返す
    
    def get_cost_of_site(self, site):
        """
        This function returns the cost to include the given site 
        to the protected area.

        Parameters
        site: string
            The ID of a site

        Return
        float
            Cost to include the site to the protected area.
        
        """

        if site not in self.costs:
            # If the site is not included in the cost data, 
            # it is assumed that the site is free from any cost.
            return 0
        else:
            return self.costs[site]


    # J: ####################################
    # J: コストの計算。指定サイトのコストの合計を返す

    def get_cost_of_sites(self, sites):
        """
        This function returns the cost to include the given sites 
        to the protected area.

        Parameters
        sites: list of strings
            The IDs of sites

        Return
        float
            Cost to include the sites to the protected area.
        """

        total_cost = 0

        for site in sites:
            if site in self.get_cost_of_site:
                total_cost += self.costs[site]

        return total_cost


    # J: ####################################
    # J: コストデータの読み込み

    def load_cost_data(self, file_name):
        """
        Load cost data from a file.

        Parameters
        file_name: string
            The name of data file
            Each line of the file consists of site ID and its cost 
            separated by a tab.

        Return
        boolean
            True (if successful) or False (if failed)
        """

        if os.path.exists(file_name) is not True:
            print("!! ", file_name, " not found in SiteCost.load_cost_data")
            return False

        fin = open(file_name, 'rt', encoding = CHR_CODE)  # read only, text file

        line = fin.readline() # Header

        while True:

            line = fin.readline()
            if not line:
                break

            data = re.split('\s+', line[:-1])  

            if len(data) < 2:   # Skip if data is incomplete
                continue

            site = data[0]         # Site ID
            cost = float(data[1])  # Cost of the site

            self.costs[site] = cost

        fin.close()

        return True
